// @dart = 2.8

import 'dart:async';
import 'dart:io';

import 'package:process_run/shell.dart';

import '../base/common.dart';
import '../base/file_system.dart';
import '../globals.dart';
import '../runner/faircli_command.dart';
import '../util/util.dart';

const dynamicTemplateProjectName = 'fair_template';
const dynamicTemplateProjectPackage = 'com.example.fair_template';
const dynamicTemplateProjectAppName = 'app_name';

const carrierTemplateProjectName = 'fair_carrier';
const carrierTemplateProjectPackage = 'com.example.fair_carrier';
const carrierTemplateProjectAppName = 'app_name';

const samplesTemplateProjectName = 'fair_samples';
const samplesTemplateProjectPackage = 'com.example.fair_samples';
const samplesTemplateProjectAppName = 'app_name';

class CreateCommand extends FaircliCommand {
  CreateCommand() {
    argParser.addOption(
      'description',
      defaultsTo: 'A new Fair project.',
      help: 'The description to use for your new Fair project.',
    );

    argParser.addOption(
      'create-kind',
      abbr: 'k',
      defaultsTo: 'dynamic',
      allowed: <String>['dynamic', 'carrier'],
    );

    argParser.addOption(
      'name',
      abbr: 'n',
      defaultsTo: '',
      help: 'create a template project',
    );

    argParser.addOption(
      'directory',
      abbr: 'd',
      defaultsTo: '',
      help: 'fair project directory',
    );

    argParser.addFlag('use-samples',
        help: 'use fair samples', defaultsTo: false);
  }

  @override
  final String name = 'create';

  @override
  final String description = 'Create a new Fair project.\n\n'
      'If run on a project that already exists, this will repair the project, recreating any files that are missing.';

  @override
  Future<FaircliCommandResult> runCommand() async {
    String projectName = argResults['name'];
    if (projectName.isEmpty) {
      throwToolExit(
          'Please use "faircli create -n xxx -k dynamic/carrier" to create a template');
    }
    var createKind = argResults['create-kind'];
    String currentPath = argResults['directory'];
    if (currentPath.isEmpty) {
      currentPath = localFileSystem.currentDirectory.path;
    }
    var projectPath = fs.path.join(currentPath, projectName);
    var projectDir = fs.directory(projectPath);
    var fileRoot = localFileSystem.currentDirectory.path;//userRootPath();
    var toolDir = fs.path.join(fileRoot, 'templates', 'tools');

    if (createKind == 'dynamic') {
      var templateDir = fs.directory(fs.path.join(
          fileRoot,
          'templates',
          boolArg('use-samples')
              ? samplesTemplateProjectName
              : dynamicTemplateProjectName));
      cliPrint(
          'Creating dynamic project from template, wait a moment please...');
      await _createProject(
          projectDir,
          templateDir,
          projectName,
          projectPath,
          toolDir,
          boolArg('use-samples')
              ? samplesTemplateProjectPackage
              : dynamicTemplateProjectPackage,
          boolArg('use-samples')
              ? samplesTemplateProjectAppName
              : dynamicTemplateProjectAppName,
          boolArg('use-samples')
              ? samplesTemplateProjectName
              : dynamicTemplateProjectName);
      cliPrint(
          'New dynamic project generated at => ${fs.path.join(currentPath, projectName)}');
    } else {
      var templateDir = fs.directory(fs.path.join(
          fileRoot,
          'templates',
          boolArg('use-samples')
              ? samplesTemplateProjectName
              : carrierTemplateProjectName));
      cliPrint(
          'Creating carrier project from template, wait a moment please...');
      await _createProject(
          projectDir,
          templateDir,
          projectName,
          projectPath,
          toolDir,
          boolArg('use-samples')
              ? samplesTemplateProjectPackage
              : carrierTemplateProjectPackage,
          boolArg('use-samples')
              ? samplesTemplateProjectAppName
              : carrierTemplateProjectAppName,
          boolArg('use-samples')
              ? samplesTemplateProjectName
              : carrierTemplateProjectName);
      cliPrint(
          'New carrier project generated at => ${fs.path.join(currentPath, projectName)}');
    }
    return null;
  }

  Future<void> _createProject(
      Directory projectDir,
      Directory templateDir,
      String projectName,
      String projectPath,
      String toolDir,
      String templateProjectPackage,
      String templateProjectAppName,
      String templateProjectName) async {
    try {
      projectDir.createSync(recursive: true);
      copyDirectory(templateDir, projectDir);
      os.chmod(projectDir, '777');
      var templateModuleMacFile =
          fs.directory(fs.path.join(projectDir.path, '__MACOSX'));
      if (templateModuleMacFile.existsSync()) {
        templateModuleMacFile.deleteSync(recursive: true);
      }
      if (platform.operatingSystem == 'windows') {
        var shell = Shell();
        await shell.runExecutableArguments(
            fs.path.join(
              toolDir,
              'project_rename.sh',
            ),
            [
              projectName,
              projectPath,
              templateProjectPackage,
              templateProjectAppName,
              templateProjectName
            ]);
      } else {
        var runResult = await Process.run(
            shellExecutable(),
            [
              'project_rename.sh',
              projectName,
              projectPath,
              templateProjectPackage,
              templateProjectAppName,
              templateProjectName
            ],
            workingDirectory: toolDir);
        if (runResult.exitCode != 0) {
          cliPrint('${runResult.stdout}');
          cliPrint('${runResult.stderr}');
          cliPrint('Rename project failed.');
        }
      }
    } catch (e) {
      throwToolExit('Failed to create project $projectDir: $e');
    }
  }
}
